﻿using System.Drawing.Imaging;

namespace TestProject;

/// <summary>
/// 색상 선택기
/// </summary>
public class ColorPicker
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 영역 - Area

    /// <summary>
    /// 영역
    /// </summary>
    public enum Area
    {
        /// <summary>
        /// 외부
        /// </summary>
        Outside,

        /// <summary>
        /// 휠
        /// </summary>
        Wheel,

        /// <summary>
        /// 삼각형
        /// </summary>
        Triangle
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Structure
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 선택 결과 - PickResult

    /// <summary>
    /// 선택 결과
    /// </summary>
    public struct PickResult
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 영역 - Area

        /// <summary>
        /// 영역
        /// </summary>
        public Area Area { get; set; }

        #endregion
        #region 색상 - Hue

        /// <summary>
        /// 색상
        /// </summary>
        public double? Hue { get; set; }

        #endregion
        #region 채도 - Saturation

        /// <summary>
        /// 채도
        /// </summary>
        public double? Saturation { get; set; }

        #endregion
        #region 값 - Value

        /// <summary>
        /// 값
        /// </summary>
        public double? Value { get; set; }

        #endregion
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Property
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 크기 - Size

    /// <summary>
    /// 크기
    /// </summary>
    public int Size { get; }

    #endregion
    #region 중앙 X - CenterX

    /// <summary>
    /// 중앙 X
    /// </summary>
    public int CenterX => Size / 2;

    #endregion
    #region 중앙 Y - CenterY

    /// <summary>
    /// 중앙 Y
    /// </summary>
    public int CenterY => Size / 2;

    #endregion
    #region 내부 반경 - InnerRadius

    /// <summary>
    /// 내부 반경
    /// </summary>
    public int InnerRadius => Size * 5 / 12;

    #endregion
    #region 외부 반경 - OuterRadius

    /// <summary>
    /// 외부 반경
    /// </summary>
    public int OuterRadius => Size / 2;

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 생성자 - ColorPicker(size)

    /// <summary>
    /// 생성자
    /// </summary>
    /// <param name="size">크기</param>
    public ColorPicker(int size = 400)
    {
        Size = size;
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 이미지 그리기 - DrawImage(hue, saturation, value)

    /// <summary>
    /// 이미지 그리기
    /// </summary>
    /// <param name="hue">색상</param>
    /// <param name="saturation">채도</param>
    /// <param name="value">값</param>
    /// <returns>이미지</returns>
    public Image DrawImage(double hue = 0d, double saturation = 1d, double value = 1d)
    {
        Bitmap bitmap = new Bitmap(Size, Size, PixelFormat.Format32bppArgb);

        for(int y = 0; y < Size; y++)
        {
            for(int x = 0; x < Size; x++)
            {
                Color color;

                PickResult pickResult = Pick(x, y);

                if(pickResult.Area == Area.Outside)
                {
                    color = Color.Transparent;
                }
                else if(pickResult.Area == Area.Wheel)
                {
                    color = GetHSVColor(pickResult.Hue.Value, saturation, value, 1);
                }
                else
                {
                    color = GetHSVColor(hue, pickResult.Saturation.Value, pickResult.Value.Value, 1);
                }

                bitmap.SetPixel(x, y, color);
            }
        }

        return bitmap;
    }

    #endregion
    #region 휠 위치 구하기 - GetWheelPosition(hue)

    /// <summary>
    /// 휠 위치 구하기
    /// </summary>
    /// <param name="hue">색상</param>
    /// <returns>휠 위치</returns>
    public Point2D GetWheelPosition(double hue)
    {
        double middleRadius = (InnerRadius + OuterRadius) / 2;

        return new Point2D
        {
            X = CenterX + middleRadius * Math.Sin(hue),
            Y = CenterY - middleRadius * Math.Cos(hue)
        };
    }

    #endregion
    #region 삼각형 위치 구하기 - GetTrianglePosition(saturation, value)

    /// <summary>
    /// 삼각형 위치 구하기
    /// </summary>
    /// <param name="saturation">채도</param>
    /// <param name="value">값</param>
    /// <returns>삼각형 위치</returns>
    public Point2D GetTrianglePosition(double saturation, double value)
    {
        double squareRoot3 = Math.Sqrt(3);

        return new Point2D
        {
            X = CenterX + InnerRadius * (2 * value - saturation * value - 1) * squareRoot3 / 2,
            Y = CenterY + InnerRadius * (1 - 3 * saturation * value) / 2
        };
    }

    #endregion
    #region 선택하기 - Pick(x, y)

    /// <summary>
    /// 선택하기
    /// </summary>
    /// <param name="x">X 좌표</param>
    /// <param name="y">Y 좌표</param>
    /// <returns>선택 결과</returns>
    public PickResult Pick(double x, double y)
    {
        double distanceFromCenter = Math.Sqrt((x - CenterX) * (x - CenterX) + (y - CenterY) * (y - CenterY));

        double squareRoot3 = Math.Sqrt(3);

        if(distanceFromCenter > OuterRadius)
        {
            return new PickResult { Area = Area.Outside };
        }
        else if(distanceFromCenter > InnerRadius)
        {
            double angle = Math.Atan2(y - CenterY, x - CenterX) + Math.PI / 2;

            if(angle < 0)
            {
                angle += 2 * Math.PI;
            }

            double hue = angle;

            return new PickResult
            {
                Area = Area.Wheel,
                Hue  = hue
            };
        }
        else
        {
            double x1 = (x - CenterX) * 1.0d / InnerRadius;
            double y1 = (y - CenterY) * 1.0d / InnerRadius;

            if(0 * x1 + 2 * y1 > 1)
            {
                return new PickResult { Area = Area.Outside };
            }
            else if(squareRoot3 * x1 + (-1) * y1 > 1)
            {
                return new PickResult { Area = Area.Outside };
            }
            else if(-squareRoot3 * x1 + (-1) * y1 > 1)
            {
                return new PickResult { Area = Area.Outside };
            }
            else
            {
                double saturation = (1 - 2 * y1) / (squareRoot3 * x1 - y1 + 2);

                double value = (squareRoot3 * x1 - y1 + 2) / 3;

                return new PickResult
                {
                    Area       = Area.Triangle,
                    Saturation = saturation,
                    Value      = value
                };
            }
        }
    }

    #endregion

    ////////////////////////////////////////////////////////////////////////////////////////// Private

    #region RGB 색상 구하기 - GetRGBColor(red, green, blue, alpha)

    /// <summary>
    /// RGB 색상 구하기
    /// </summary>
    /// <param name="red">적색</param>
    /// <param name="green">녹색</param>
    /// <param name="blue">청색</param>
    /// <param name="alpha">투명도</param>
    /// <returns>RGB 색상</returns>
    private Color GetRGBColor(double red, double green, double blue, double alpha)
    {
        return Color.FromArgb
        (
            Math.Min(255, (int)(alpha * 256)),
            Math.Min(255, (int)(red   * 256)),
            Math.Min(255, (int)(green * 256)),
            Math.Min(255, (int)(blue  * 256))
        );
    }

    #endregion
    #region HSV 색상 구하기 - GetHSVColor(hue, saturation, value, alpha)

    /// <summary>
    /// HSV 색상 구하기
    /// </summary>
    /// <param name="hue">색상</param>
    /// <param name="saturation">채도</param>
    /// <param name="value">값</param>
    /// <param name="alpha">알파</param>
    /// <returns>색상</returns>
    private Color GetHSVColor(double hue, double saturation, double value, double alpha)
    {
        double chroma = value * saturation;
        double step   = Math.PI / 3;
        double interm = chroma * (1 - Math.Abs((hue / step) % 2.0 - 1));
        double shift  = value - chroma;

        if(hue < 1 * step) return GetRGBColor(shift + chroma, shift + interm, shift + 0     , alpha);
        if(hue < 2 * step) return GetRGBColor(shift + interm, shift + chroma, shift + 0     , alpha);
        if(hue < 3 * step) return GetRGBColor(shift + 0     , shift + chroma, shift + interm, alpha);
        if(hue < 4 * step) return GetRGBColor(shift + 0     , shift + interm, shift + chroma, alpha);
        if(hue < 5 * step) return GetRGBColor(shift + interm, shift + 0     , shift + chroma, alpha);

        return GetRGBColor(shift + chroma, shift + 0, shift + interm, alpha);
    }

    #endregion
}